/*
 * i8253.c  8253/PIT functions
 *
 */
#include <linux/clockchips.h>
#include <linux/spinlock.h>
#include <linux/jiffies.h>
#include <linux/sysdev.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/mca.h>

#include <asm/smp.h>
#include <asm/io_apic.h>
#include <asm/delay.h>
#include <asm/i8253.h>
#include <asm/io.h>

#include "io_ports.h"

DEFINE_SPINLOCK(i8253_lock);
EXPORT_SYMBOL(i8253_lock);

static void init_pit_timer(int mode)
{
	unsigned long flags;

	spin_lock_irqsave(&i8253_lock, flags);

	if (mode != CLOCK_EVT_ONESHOT) {
		/* binary, mode 2, LSB/MSB, ch 0 */
		outb_p(0x34, PIT_MODE);
		udelay(10);
		outb_p(LATCH & 0xff , PIT_CH0);	/* LSB */
		outb(LATCH >> 8 , PIT_CH0);	/* MSB */
	} else {
		/* One shot setup */
		outb_p(0x38, PIT_MODE);
		udelay(10);
	}

	spin_unlock_irqrestore(&i8253_lock, flags);
}

static void pit_next_event(unsigned long evt)
{
	unsigned long flags;

	spin_lock_irqsave(&i8253_lock, flags);
	outb_p(evt & 0xff , PIT_CH0);	/* LSB */
	outb(evt >> 8 , PIT_CH0);	/* MSB */
	spin_unlock_irqrestore(&i8253_lock, flags);
}

static struct clock_event pit_clockevent = {
	.name		= "pit",
	.capabilities	= CLOCK_CAP_TICK
#ifndef CONFIG_SMP
			| CLOCK_CAP_NEXTEVT | CLOCK_CAP_PROFILE |
			CLOCK_CAP_UPDATE
#endif
	,
	.set_mode	= init_pit_timer,
	.set_next_event = pit_next_event,
	.start_event	= io_apic_timer_ack,
	.end_event	= mca_timer_ack,
	.shift		= 32,
	.irq		= 0,
};

void setup_pit_timer(void)
{
	pit_clockevent.mult = div_sc32(CLOCK_TICK_RATE, NSEC_PER_SEC);
	pit_clockevent.max_delta_ns =
		clockevent_delta2ns(0x7FFF, &pit_clockevent);
	pit_clockevent.min_delta_ns =
		clockevent_delta2ns(0xF, &pit_clockevent);
	setup_global_clockevent(&pit_clockevent, CPU_MASK_NONE);
}

/*
 * Since the PIT overflows every tick, its not very useful
 * to just read by itself. So use jiffies to emulate a free
 * running counter:
 */
static cycle_t pit_read(void)
{
	unsigned long flags, seq;
	int count;
	u64 jifs;

	do {
		seq = read_seqbegin(&xtime_lock);

		spin_lock_irqsave(&i8253_lock, flags);
		outb_p(0x00, PIT_MODE);	/* latch the count ASAP */
		count = inb_p(PIT_CH0);	/* read the latched count */
		count |= inb_p(PIT_CH0) << 8;

		/* VIA686a test code... reset the latch if count > max + 1 */
		if (count > LATCH) {
			outb_p(0x34, PIT_MODE);
			outb_p(LATCH & 0xff, PIT_CH0);
			outb(LATCH >> 8, PIT_CH0);
			count = LATCH - 1;
		}
		spin_unlock_irqrestore(&i8253_lock, flags);

		jifs = jiffies_64;
	} while (read_seqretry(&xtime_lock, seq));

	jifs -= INITIAL_JIFFIES;
	count = (LATCH-1) - count;

	return (cycle_t)(jifs * LATCH) + count;
}

static struct clocksource clocksource_pit = {
	.name	= "pit",
	.rating = 110,
	.read	= pit_read,
	.mask	= (cycle_t)-1,
	.mult	= 0,
	.shift	= 20,
};

static int __init init_pit_clocksource(void)
{
	if (num_possible_cpus() > 4) /* PIT does not scale! */
		return 0;

	clocksource_pit.mult = clocksource_hz2mult(CLOCK_TICK_RATE, 20);
	register_clocksource(&clocksource_pit);

	return 0;
}
module_init(init_pit_clocksource);
