/*
 * linux/kernel/time/clockevents.c
 *
 * This file contains functions which manage clock event drivers.
 *
 * Copyright(C) 2005 Thomas Gleixner <tglx@linutronix.de>
 *
 * Kudos to Ingo Molnar for review, criticism, ideas
 *
 * We have two types of clock event devices:
 * - global events (one device per system)
 * - local events (one device per cpu)
 *
 * We assign the various time(r) related interrupts to those devices
 *
 * - global tick
 * - profiling (per cpu)
 * - next timer events (per cpu)
 *
 * TODO:
 * - implement variable frequency profiling
 */

#include <linux/clockchips.h>
#include <linux/cpu.h>
#include <linux/init.h>
#include <linux/notifier.h>
#include <linux/module.h>
#include <linux/percpu.h>
#include <linux/sysdev.h>
#include <linux/hrtimer.h>

#define MAX_CLOCK_EVENTS	4

struct event_descr {
	struct clock_event *event;
	unsigned int mode;
	unsigned int real_caps;
	struct irqaction action;
};

struct local_events {
	int installed;
	struct event_descr events[MAX_CLOCK_EVENTS];
	struct clock_event *nextevt;
};

/* Variables related to the global event source */
static struct event_descr global_eventsource;

/* Variables related to the per cpu local event sources */
static DEFINE_PER_CPU(struct local_events, local_eventsources);

#ifdef CONFIG_SMP
# define recalc_global_event(e) do { } while(0)
#else
# define recalc_global_event(c) recalc_active_event(&global_eventsource, c)
#endif

/*
 * Math helper. Convert a latch value to ns
 */
unsigned long clockevent_delta2ns(unsigned long latch, struct clock_event *evt)
{
	u64 clc = ((u64) latch << evt->shift);

	do_div(clc, evt->mult);
	if (clc < KTIME_MONOTONIC_RES.tv64)
		clc = KTIME_MONOTONIC_RES.tv64;
	if (clc > 0x7FFFFFFF)
		clc = 0x7FFFFFFF;

	return (unsigned long) clc;
}

/*
 * Generic timer interrupt handler usable for all kinds of events
 */
static irqreturn_t timer_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct clock_event *evt = dev_id;

	if (evt->start_event)
		evt->start_event(evt->priv);

	evt->event_handler(regs);

	if (evt->end_event)
		evt->end_event(evt->priv);

	return IRQ_HANDLED;
}

/*
 * Handle tick
 */
static void handle_tick(struct pt_regs *regs)
{
	write_seqlock(&xtime_lock);
	do_timer(regs);
	write_sequnlock(&xtime_lock);
}

/*
 * Handle tick and update
 */
static void handle_tick_update(struct pt_regs *regs)
{
	write_seqlock(&xtime_lock);
	do_timer(regs);
	write_sequnlock(&xtime_lock);

	update_process_times(user_mode(regs));
}

/*
 * Handle tick, update and profiling
 */
static void handle_tick_update_profile(struct pt_regs *regs)
{
	write_seqlock(&xtime_lock);
	do_timer(regs);
	write_sequnlock(&xtime_lock);

	update_process_times(user_mode(regs));
	profile_tick(CPU_PROFILING, regs);
}

/*
 * Handle update
 */
static void handle_update(struct pt_regs *regs)
{
	update_process_times(user_mode(regs));
}

/*
 * Handle update and profile
 */
static void handle_update_profile(struct pt_regs *regs)
{
	update_process_times(user_mode(regs));
	profile_tick(CPU_PROFILING, regs);
}

/*
 * Handle profile
 */
static void handle_profile(struct pt_regs *regs)
{
	profile_tick(CPU_PROFILING, regs);
}

#ifdef CONFIG_HIGH_RES_TIMERS
/*
 * Handle next event
 */
static void handle_nextevent(struct pt_regs *regs)
{
	hrtimer_interrupt();
}

/*
 * Handle next event, tick
 */
static void handle_nextevent_tick(struct pt_regs *regs)
{
	int res;

	res = hrtimer_interrupt();
	for (; res > 0; res--)
		handle_tick(regs);
}

/*
 * Handle next event, update
 */
static void handle_nextevent_update(struct pt_regs *regs)
{
	if (hrtimer_interrupt() > 0)
		handle_update(regs);
}

/*
 * Handle next event, tick, update
 */
static void handle_nextevent_tick_update(struct pt_regs *regs)
{
	int res;

	if ((res = hrtimer_interrupt()) == 0)
		return;

	for (; res > 0; res--)
		handle_tick(regs);

	handle_update(regs);
}

/*
 * Handle next event, profile
 */
static void handle_nextevent_profile(struct pt_regs *regs)
{
	if (hrtimer_interrupt() > 0)
		handle_profile(regs);
}

/*
 * Handle next event, update, profile
 */
static void handle_nextevent_update_profile(struct pt_regs *regs)
{
	if (hrtimer_interrupt() > 0)
		handle_update_profile(regs);
}

/*
 * Handle next event, tick, update, profile
 */
static void handle_nextevent_all(struct pt_regs *regs)
{
	int res;

	if ((res = hrtimer_interrupt()) == 0)
		return;

	for (; res > 0; res--)
		handle_tick(regs);

	handle_update_profile(regs);
}
#else
#define handle_nextevent		NULL
#define handle_nextevent_tick		NULL
#define handle_nextevent_update		NULL
#define handle_nextevent_tick_update	NULL
#define handle_nextevent_profile	NULL
#define handle_nextevent_update_profile	NULL
#define handle_nextevent_all		NULL
#endif

/*
 * Lookup table for event assignment
 */
static void *event_handlers[] = {
	NULL,				/* 0: No capability selected */
	handle_tick,			/* 1: Tick only	*/
	handle_nextevent,		/* 2: Next event only */
	handle_nextevent_tick,		/* 3: Next event + tick */
	handle_update,			/* 4: Update process times */
	handle_tick_update,		/* 5: Tick + update process times */
	handle_nextevent_update,	/* 6: Next event +
					      update process times */
	handle_nextevent_tick_update,	/* 7: Next event + tick +
					      update process times */
	handle_profile,			/* 8: Profiling int */
	NULL,				/* 9: Tick + profiling */
	handle_nextevent_profile,	/* A: Next event + profiling */
	NULL,				/* B: Next event + tick + profiling */
	handle_update_profile,		/* C: Update process times +
					      profiling */
	handle_tick_update_profile,	/* D: Tick + update process times +
					      profiling */
	handle_nextevent_update_profile,/* E: Next event +
					      update process times +
					      profiling */
	handle_nextevent_all,		/* F: Next event + tick +
					      update process times +
					      profiling */
};

/*
 * The selection model makes following assumptions:
 *
 * There is only one global event source set up. Global event sources
 * are unique devices in a system (UP/SMP) Usually they are setup
 * early in the bootup phase to provide the basic tick environment to
 * bring up hardware. Such a device can be capable of providing all in
 * one functionality including next event scheduling.
 *
 * When a system has decicated event sources which can be used for
 * particular purposes then we assume that there are no devices setup
 * which provide "competing" functionality. i.e. the developer has to
 * decide which device should be used for a particular functionality
 * rather than letting the management code guess about the best
 * fit. The code manages the cases, where the number of event sources
 * is unknown during compile time, but the functionality of the event
 * source is assigned to the respective event source by a human best
 * fit decision.
 *
 * The purpose of the management code is to provide handling code for
 * the various possible combinations and the necessary infrastructure
 * to handle next event (e.g. high resolution) scheduling with a
 * single event source, which makes a periodic rescheduling of the
 * tick interupt necessary. This is done to avoid the #ifdef mess all
 * over the architecture dependend timer and event interupt code for
 * the various possible use case combinations and allows clean non
 * intrusive implementation of configurable extensions to the time
 * related event system e.g. dynamic ticks, high resolution
 * timers.
 *
 * Some architectures can use a NMI based profiling mechanism. If this
 * is used, then profiling is excluded from the event assignements.
 *
 * SMP systems CPU which have no unique global event source should not
 * setup a global event source. The correct way is setting up one
 * event source (usually local to CPU0 or the bootcpu in hotplug
 * systems) which has the CLOCK_CAP_TICK flag set, so the management
 * code assigns exactly one tick source for the complete system.
 *
 * A special case are pseudo event sources (IPI mechanisms) on SMP
 * systems. They can be used for populating tick events from one event
 * source across multiple CPUs.
 *
 */
static int setup_event(struct event_descr *descr, struct clock_event *evt,
		       unsigned int caps, cpumask_t cpumask)
{
	void *handler = event_handlers[caps];

	if (!handler) {
		printk(KERN_ERR "Unsupported event source %s\n", evt->name);
		return -EINVAL;
	}

	/* Store the event handler */
	evt->event_handler = handler;

	/* Save the event descriptor reference */
	descr->event = evt;

	if (!(evt->capabilities & CLOCK_HAS_IRQHANDLER)) {
		descr->action.name = evt->name;
		descr->action.handler = timer_interrupt;
		descr->action.flags = SA_INTERRUPT;
		descr->action.mask = cpumask;
		descr->action.dev_id = evt;
		setup_irq(evt->irq, &descr->action);
	}

	descr->real_caps = caps;
	descr->mode = CLOCK_EVT_STARTUP;
	if (evt->set_mode)
		evt->set_mode(CLOCK_EVT_STARTUP);
	printk(KERN_INFO "Event source %s installed with caps set: %02x\n",
	       descr->event->name, descr->real_caps);

	return 0;
}

/*
 * Mask out the functionality which is covered by the new event source
 * and assign a new event handler.
 */
static unsigned int recalc_active_event(struct event_descr *descr,
					unsigned int caps)
{
	unsigned int gcaps;

	if (!descr->event)
		return caps;

	/* Find out the overlapping bits */
	gcaps = descr->real_caps & caps;

	/*
	 * Be careful here. We dont know in which order the event
	 * sources are set up. So we might switch off a previously
	 * registered source completely.
	 *
	 * Might need more thoughts though.
	 */
	if (gcaps == descr->real_caps) {
		int i;

		i = ffs(gcaps) - 1;
		gcaps &= ~(1 << i);
		caps &= ~(1 << i);
	}
	if (!gcaps)
		return caps;

	/* Mask the bits which are now covered by the new event */
	descr->real_caps &= ~gcaps;

	/* Assign the new event handler */
	descr->event->event_handler = event_handlers[descr->real_caps];
	printk(KERN_INFO "Event source %s new caps set: %02x\n" ,
	       descr->event->name, descr->real_caps);

	return caps;
}

/*
 * Recalc the events and reassign the handlers if necessary
 */
static int recalc_events(struct local_events *sources, struct clock_event *evt,
			 cpumask_t cpumask)
{
	unsigned int caps = evt->capabilities & CLOCK_CAP_MASK;
	int i;

	if (sources->installed == MAX_CLOCK_EVENTS)
		return -ENOSPC;

	if (!event_handlers[caps])
		return -EINVAL;

	recalc_global_event(caps);

	for (i = 0; i < sources->installed; i++)
		caps = recalc_active_event(&sources->events[i], caps);

	setup_event(&sources->events[sources->installed], evt, caps, cpumask);
	sources->installed++;
	if (evt->capabilities & CLOCK_CAP_NEXTEVT) {
		sources->nextevt = evt;
		hrtimer_clock_notify();
	}

	return 0;
}

/**
 * setup_local_clockevent - Set up a cpu local clock event device
 *
 * @evtdev:	event device to be registered
 * @cpumask:	cpumask for the irq setup
 */
int setup_local_clockevent(struct clock_event *evtdev, cpumask_t cpumask)
{
	struct local_events *sources = &__get_cpu_var(local_eventsources);
	unsigned long flags;
	int res;

	/* Recalc event sources and maybe reassign interrupts */
	local_irq_save(flags);
	res = recalc_events(sources, evtdev, cpumask);
	local_irq_restore(flags);

	return res;
}
EXPORT_SYMBOL(setup_local_clockevent);

/**
 * set_global_clockevent - Set the device which generates global clock events
 *
 * @evt:	The device which generates global clock events (ticks)
 */
int __init setup_global_clockevent(struct clock_event *evt, cpumask_t cpumask)
{
	int res;

	res = setup_event(&global_eventsource, evt,
			   evt->capabilities & CLOCK_CAP_MASK, cpumask);
#ifndef CONFIG_SMP
	/*
	 * The "global" event source on UP systems can serve as
	 * next event source !
	 */
	if (!res && (evt->capabilities & CLOCK_CAP_NEXTEVT))
		per_cpu(local_eventsources, 0).nextevt = evt;
#endif
	return res;
}

/**
 * clockevents_next_event_available - Check for a installed next event source
 */
int clockevents_next_event_available(void)
{
	struct local_events *sources = &__get_cpu_var(local_eventsources);
        int i;

	if (!sources->nextevt)
		return 0;

#ifndef CONFIG_SMP
	if (sources->nextevt == global_eventsource.event)
		return CLOCK_EVT_SCHEDTICK;
#endif
	/*
	 * Check, whether the next event source is solely for next events or
	 * it has to do some periodic tick functionality
	 * We use the real_caps field here, as some other source might
	 * have switched off one of the capability flags.
	 */
	for (i = 0; i < sources->installed; i++) {
                if (sources->nextevt != sources->events[i].event)
                        continue;

	        if (sources->events[i].real_caps & ~CLOCK_CAP_NEXTEVT)
        		return CLOCK_EVT_SCHEDTICK;
		return CLOCK_EVT_NOTICK;
        }
	return CLOCK_EVT_NOTICK;
}

int clockevents_init_next_event(void)
{
	struct local_events *sources = &__get_cpu_var(local_eventsources);

	if (!sources->nextevt)
		return 0;

	if (sources->nextevt->set_mode)
		sources->nextevt->set_mode(CLOCK_EVT_ONESHOT);

	return 1;
}

int clockevents_set_next_event(ktime_t expires)
{
	struct local_events *sources = &__get_cpu_var(local_eventsources);
	nsec_t delta = ktime_to_ns(ktime_sub(expires, ktime_get()));
	unsigned long clc;

	if (delta <= 0)
		return -ETIME;
	if (delta > sources->nextevt->max_delta_ns)
		delta = sources->nextevt->max_delta_ns;
	if (delta < sources->nextevt->min_delta_ns)
		delta = sources->nextevt->min_delta_ns;

	clc = mpy_sc32((unsigned long) delta, sources->nextevt->mult);
	sources->nextevt->set_next_event(clc);
	return 0;
}

void clockevents_trigger_next_event(void)
{
}

#ifdef CONFIG_PM
static int
global_eventsource_suspend(struct sys_device *dev, pm_message_t state)
{
	/* Do generic stuff here */
	if (global_eventsource.event->suspend)
		global_eventsource.event->suspend();
	return 0;
}

static int global_eventsource_resume(struct sys_device *dev)
{
	/* Do generic stuff here */
	if (global_eventsource.event->resume)
		global_eventsource.event->resume();
	return 0;
}
#else
# define global_eventsource_resume	NULL
# define global_eventsource_suspend	NULL
#endif

static struct sysdev_class global_clock_event_sysclass = {
	.resume = global_eventsource_resume,
	.suspend = global_eventsource_suspend,
	set_kset_name("global_clock_event"),
};

static struct sys_device device_global_clock_event = {
	.id	= 0,
	.cls	= &global_clock_event_sysclass,
};

static int __init global_clock_event_devinit(void)
{
	int error = sysdev_class_register(&global_clock_event_sysclass);

	if (!error)
		error = sysdev_register(&device_global_clock_event);

	return error;
}

device_initcall(global_clock_event_devinit);

/*
 * Functions related to initialization
 */
static void __devinit init_clockevents_cpu(int cpu)
{
}

static int __devinit clockevents_cpu_notify(struct notifier_block *self,
					    unsigned long action, void *hcpu)
{
	long cpu = (long)hcpu;

	switch(action) {
	case CPU_UP_PREPARE:
		init_clockevents_cpu(cpu);
		break;
#ifdef CONFIG_HOTPLUG_CPU
	case CPU_DEAD:
		break;
#endif
	default:
		break;
	}
	return NOTIFY_OK;
}

static struct notifier_block __devinitdata clockevents_nb = {
	.notifier_call	= clockevents_cpu_notify,
};

void __init init_clockevents(void)
{
	clockevents_cpu_notify(&clockevents_nb, (unsigned long)CPU_UP_PREPARE,
				(void *)(long)smp_processor_id());
	register_cpu_notifier(&clockevents_nb);
}
