/*
 * linux/kernel/time/timeofday.c
 *
 * This file contains the functions which access and manage
 * the system's time of day functionality.
 *
 * Copyright (C) 2003, 2004, 2005 IBM, John Stultz (johnstul@us.ibm.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * TODO WishList:
 *   o See XXX's below.
 */

#include <linux/clocksource.h>
#include <linux/timeofday.h>
#include <linux/jiffies.h>
#include <linux/sysdev.h>
#include <linux/timer.h>
#include <linux/module.h>
#include <linux/ktime.h>
#include <linux/timex.h>
#include <linux/sched.h>

#include <asm/timeofday.h>

/* Periodic hook interval */
#define PERIODIC_INTERVAL_MS 50

/* [ktime_t based variables]
 * system_time:
 *	Monotonically increasing counter of the number of nanoseconds
 *	since boot.
 * wall_time_offset:
 *	Offset added to system_time to provide accurate time-of-day
 */
static ktime_t system_time;
static ktime_t wall_time_offset;

/* [timespec based variables]
 * These variables mirror teh ktime_t based variables to avoid
 * performance issues in the userspace syscall paths.
 *
 * wall_time_ts:
 *	timespec holding the current wall time.
 *	NOTE: clock_was_set() must be called when this value changes.
 * mono_time_ts:
 *	timespec holding the current monotonic time.
 * monotonic_time_offset_ts:
 *	timespec holding the difference between wall and monotonic time.
 */
static struct timespec wall_time_ts;
static struct timespec mono_time_ts;
static struct timespec monotonic_time_offset_ts;

/* [cycle based variables]
 * cycle_last:
 *	Value of the clocksource at the last timeofday_periodic_hook()
 *	(adjusted only minorly to account for rounded off cycles)
 */
static cycle_t cycle_last;

/* [clocksource_interval variables]
 * ts_interval:
 *	This clocksource_interval is used in the fixed interval
 *	cycles to nanosecond calculation.
 * INTERVAL_LEN:
 *	This constant is the requested fixed interval period
 *	in nanoseconds.
 */
struct clocksource_interval ts_interval;
#define INTERVAL_LEN ((PERIODIC_INTERVAL_MS-1)*1000000)

/* [clocksource data]
 * clock:
 *	current clocksource pointer
 */
static struct clocksource *clock;

/* [NTP adjustment]
 * ntp_adj:
 *	value of the current ntp adjustment, stored in
 *	clocksource multiplier units.
 */
static int ntp_adj;

/* [locks]
 * system_time_lock:
 *	generic lock for all locally scoped time values
 */
static seqlock_t system_time_lock = SEQLOCK_UNLOCKED;


/* [suspend/resume info]
 * time_suspend_state:
 *	variable that keeps track of suspend state
 * suspend_start:
 *	start of the suspend call
 */
static enum {
	TIME_RUNNING,
	TIME_SUSPENDED
} time_suspend_state = TIME_RUNNING;

static nsec_t suspend_start;

/* [Soft-Timers]
 * timeofday_timer:
 *	soft-timer used to call timeofday_periodic_hook()
 */
static struct timer_list timeofday_timer;

/**
 * update_legacy_time_values - sync legacy time values
 *
 * This function is necessary for a smooth transition to the
 * new timekeeping code. When all the xtime/wall_to_monotonic
 * users are converted this function can be removed.
 *
 * system_time_lock must be held by the caller
 */
static void update_legacy_time_values(void)
{
	unsigned long flags;

	write_seqlock_irqsave(&xtime_lock, flags);

	xtime = wall_time_ts;
	set_normalized_timespec(&wall_to_monotonic,
		-monotonic_time_offset_ts.tv_sec,
		-monotonic_time_offset_ts.tv_nsec);

	write_sequnlock_irqrestore(&xtime_lock, flags);

	/* since time state has changed, notify vsyscall code */
	arch_update_vsyscall_gtod(wall_time_ts, cycle_last, clock, ntp_adj);
}

/**
 * __get_nsec_offset - Returns nanoseconds since last call to periodic_hook
 *
 * private function, must hold system_time_lock lock when being
 * called. Returns the number of nanoseconds since the
 * last call to timeofday_periodic_hook() (adjusted by NTP scaling)
 */
static inline nsec_t __get_nsec_offset(void)
{
	cycle_t cycle_now, cycle_delta;
	nsec_t ns_offset;

	/* read clocksource: */
	cycle_now = read_clocksource(clock);

	/* calculate the delta since the last timeofday_periodic_hook: */
	cycle_delta = (cycle_now - cycle_last) & clock->mask;

	/* convert to nanoseconds: */
	ns_offset = cyc2ns(clock, ntp_adj, cycle_delta);

	/*
	 * special case for jiffies tick/offset based systems,
	 * add arch-specific offset:
	 */
	ns_offset += arch_getoffset();

	return ns_offset;
}

/**
 * __get_monotonic_clock - Returns monotonically increasing nanoseconds
 *
 * private function, must hold system_time_lock lock when being
 * called. Returns the monotonically increasing number of
 * nanoseconds since the system booted (adjusted by NTP scaling)
 */
static ktime_t __get_monotonic_clock(void)
{
	nsec_t offset = __get_nsec_offset();
	return ktime_add_ns(system_time, offset);
}

/**
 * get_monotonic_clock - Returns monotonic time in ktime_t format
 *
 * Returns the monotonically increasing number of nanoseconds
 * since the system booted via __monotonic_clock()
 */
ktime_t get_monotonic_clock(void)
{
	unsigned long seq;
	ktime_t ret;

	/* atomically read __get_monotonic_clock() */
	do {
		seq = read_seqbegin(&system_time_lock);

		ret = __get_monotonic_clock();

	} while (read_seqretry(&system_time_lock, seq));

	return ret;
}

EXPORT_SYMBOL_GPL(get_monotonic_clock);

/**
 * get_realtime_clock - Returns the timeofday in ktime_t format
 *
 * Returns the wall time in ktime_t format. The resolution
 * is nanoseconds
 */
ktime_t get_realtime_clock(void)
{
	unsigned long seq;
	ktime_t ret;

	/* atomically read __get_monotonic_clock() */
	do {
		seq = read_seqbegin(&system_time_lock);

		ret = __get_monotonic_clock();
		ret = ktime_add(ret, wall_time_offset);

	} while (read_seqretry(&system_time_lock, seq));

	return ret;
}

/**
 * get_realtime_offset - Returns the offset of realtime clock
 *
 * Returns the number of nanoseconds in ktime_t storage format which
 * represents the offset of the realtime clock to the the monotonic clock
 */
ktime_t get_realtime_offset(void)
{
	unsigned long seq;
	ktime_t ret;

	/* atomically read wall_time_offset */
	do {
		seq = read_seqbegin(&system_time_lock);

		ret = wall_time_offset;

	} while (read_seqretry(&system_time_lock, seq));

	return ret;
}

/**
 * get_monotonic_clock_ts - Returns monotonic time in timespec format
 * @ts:		pointer to the timespec to be set
 *
 * Returns a timespec of nanoseconds since the system booted and
 * store the result in the timespec variable pointed to by @ts
 */
void get_monotonic_clock_ts(struct timespec *ts)
{
	unsigned long seq;
	nsec_t offset;

	do {
		seq = read_seqbegin(&system_time_lock);

		*ts = mono_time_ts;
		offset = __get_nsec_offset();
	} while (read_seqretry(&system_time_lock, seq));

	timespec_add_ns(ts, offset);
}

/**
 * __get_realtime_clock_ts - Returns the time of day in a timespec
 * @ts:		pointer to the timespec to be set
 *
 * Returns the time of day in a timespec. Used by
 * do_gettimeofday() and get_realtime_clock_ts().
 */
static inline void __get_realtime_clock_ts(struct timespec *ts)
{
	unsigned long seq;
	nsec_t nsecs;

	do {
		seq = read_seqbegin(&system_time_lock);

		*ts = wall_time_ts;
		nsecs = __get_nsec_offset();

	} while (read_seqretry(&system_time_lock, seq));

	timespec_add_ns(ts, nsecs);
}

/**
 * get_realtime_clock_ts - Returns the time of day in a timespec
 * @ts:		pointer to the timespec to be set
 *
 * Returns the time of day in a timespec.
 */
void get_realtime_clock_ts(struct timespec *ts)
{
	__get_realtime_clock_ts(ts);
}

EXPORT_SYMBOL(get_realtime_clock_ts);

/**
 * do_gettimeofday - Returns the time of day in a timeval
 * @tv:		pointer to the timeval to be set
 *
 * NOTE: Users should be converted to using get_realtime_clock_ts()
 */
void do_gettimeofday(struct timeval *tv)
{
	struct timespec now;

	__get_realtime_clock_ts(&now);
	tv->tv_sec = now.tv_sec;
	tv->tv_usec = now.tv_nsec/1000;
}

EXPORT_SYMBOL(do_gettimeofday);

/**
 * do_settimeofday - Sets the time of day
 * @tv:		pointer to the timespec variable containing the new time
 *
 * Sets the time of day to the new time and update NTP and notify hrtimers
 */
int do_settimeofday(struct timespec *tv)
{
	unsigned long flags;
	ktime_t newtime;

	newtime = timespec_to_ktime(*tv);

	write_seqlock_irqsave(&system_time_lock, flags);

	/* calculate the new offset from the monotonic clock */
	wall_time_offset = ktime_sub(newtime, __get_monotonic_clock());

	/* update the internal timespec variables */
	wall_time_ts = ktime_to_timespec(ktime_add(system_time,
						wall_time_offset));
	monotonic_time_offset_ts = ktime_to_timespec(wall_time_offset);

	ntp_clear();
	update_legacy_time_values();

	write_sequnlock_irqrestore(&system_time_lock, flags);

	/* signal hrtimers about time change */
	clock_was_set();

	return 0;
}

EXPORT_SYMBOL(do_settimeofday);

/**
 * __increment_system_time - Increments system time
 * @delta:	nanosecond delta to add to the time variables
 *
 * Private helper that increments system_time and related
 * timekeeping variables.
 */
static void __increment_system_time(nsec_t delta)
{
	system_time = ktime_add_ns(system_time, delta);
	timespec_add_ns(&wall_time_ts, delta);
	timespec_add_ns(&mono_time_ts, delta);
}

/**
 * timeofday_suspend_hook - allows the timeofday subsystem to be shutdown
 * @dev:	unused
 * @state:	unused
 *
 * This function allows the timeofday subsystem to be shutdown for a period
 * of time. Called when going into suspend/hibernate mode.
 */
static int timeofday_suspend_hook(struct sys_device *dev, pm_message_t state)
{
	unsigned long flags;

	write_seqlock_irqsave(&system_time_lock, flags);

	BUG_ON(time_suspend_state != TIME_RUNNING);

	/*
	 * First off, save suspend start time
	 * then quickly accumulate the current nsec offset.
	 * These two calls hopefully occur quickly
	 * because the difference between reads will
	 * accumulate as time drift on resume.
	 */
	suspend_start = read_persistent_clock();
	__increment_system_time(__get_nsec_offset());

	time_suspend_state = TIME_SUSPENDED;

	write_sequnlock_irqrestore(&system_time_lock, flags);

	return 0;
}

/**
 * timeofday_resume_hook - Resumes the timeofday subsystem.
 * @dev:	unused
 *
 * This function resumes the timeofday subsystem from a previous call
 * to timeofday_suspend_hook.
 */
static int timeofday_resume_hook(struct sys_device *dev)
{
	nsec_t suspend_end, suspend_time;
	unsigned long flags;

	write_seqlock_irqsave(&system_time_lock, flags);

	BUG_ON(time_suspend_state != TIME_SUSPENDED);

	/*
	 * Read persistent clock to mark the end of
	 * the suspend interval then rebase the
	 * cycle_last to current clocksource value.
	 * Again, time between these two calls will
	 * not be accounted for and will show up as
	 * time drift.
	 */
	suspend_end = read_persistent_clock();
	cycle_last = read_clocksource(clock);

	/* calculate suspend time and add it to system time: */
	suspend_time = suspend_end - suspend_start;
	__increment_system_time(suspend_time);

	ntp_clear();

	time_suspend_state = TIME_RUNNING;

	update_legacy_time_values();

	write_sequnlock_irqrestore(&system_time_lock, flags);

	/* notify the posix timers if wall_time_offset changed */
	clock_was_set();

	return 0;
}

/* sysfs resume/suspend bits */
static struct sysdev_class timeofday_sysclass = {
	.resume		= timeofday_resume_hook,
	.suspend	= timeofday_suspend_hook,
	set_kset_name("timeofday"),
};

static struct sys_device device_timer = {
	.id		= 0,
	.cls		= &timeofday_sysclass,
};

static int timeofday_init_device(void)
{
	int error = sysdev_class_register(&timeofday_sysclass);

	if (!error)
		error = sysdev_register(&device_timer);

	return error;
}

device_initcall(timeofday_init_device);

/**
 * timeofday_periodic_hook - Does periodic update of timekeeping values.
 * @unused:	unused value
 *
 * Calculates the delta since the last call, updates system time and
 * clears the offset.
 *
 * Called via timeofday_timer.
 */
static void timeofday_periodic_hook(unsigned long unused)
{
	unsigned long flags;

	cycle_t cycle_now, cycle_delta;
	nsec_t delta_nsec;
	static u64 remainder;

	long leapsecond = 0;
	struct clocksource* next;

	int ppm;
	static int ppm_last;

	int something_changed = 0, clocksource_changed = 0;
	struct clocksource old_clock;
	static nsec_t second_check;

	write_seqlock_irqsave(&system_time_lock, flags);

	/* read time source & calc time since last call: */
	cycle_now = read_clocksource(clock);
	cycle_delta = (cycle_now - cycle_last) & clock->mask;

	delta_nsec = cyc2ns_fixed_rem(ts_interval, &cycle_delta, &remainder);
	cycle_last = (cycle_now - cycle_delta)&clock->mask;

	/* update system_time:  */
	__increment_system_time(delta_nsec);

	/* advance the ntp state machine by ns interval: */
	ntp_advance(delta_nsec);

	/* only call ntp_leapsecond and ntp_sync once a sec:  */
	second_check += delta_nsec;
	if (second_check >= NSEC_PER_SEC) {
		/* do ntp leap second processing: */
		leapsecond = ntp_leapsecond(wall_time_ts);
		if (leapsecond) {
			wall_time_offset = ktime_add_ns(wall_time_offset,
						leapsecond * NSEC_PER_SEC);
			wall_time_ts.tv_sec += leapsecond;
			monotonic_time_offset_ts.tv_sec += leapsecond;
		}
		/* sync the persistent clock: */
		if (ntp_synced())
			sync_persistent_clock(wall_time_ts);
		second_check -= NSEC_PER_SEC;
	}

	/* if necessary, switch clocksources: */
	next = get_next_clocksource();
	if (next != clock) {
		/* immediately set new cycle_last: */
		cycle_last = read_clocksource(next);
		/* update cycle_now to avoid problems in accumulation later: */
		cycle_now = cycle_last;
		/* swap clocksources: */
		old_clock = *clock;
		clock = next;
		printk(KERN_INFO "Time: %s clocksource has been installed.\n",
					clock->name);
		ntp_clear();
		ntp_adj = 0;
		remainder = 0;
		something_changed = 1;
		clocksource_changed = 1;
	}

	/*
	 * now is a safe time, so allow clocksource to adjust
	 * itself (for example: to make cpufreq changes):
	 */
	if (clock->update_callback) {
		/*
		 * since clocksource state might change,
		 * keep a copy, but only if we've not
		 * already changed timesources:
		 */
		if (!something_changed)
			old_clock = *clock;
		if (clock->update_callback()) {
			remainder = 0;
			something_changed = 1;
		}
	}

	/* check for new PPM adjustment: */
	ppm = ntp_get_ppm_adjustment();
	if (ppm_last != ppm) {
		/* make sure old_clock is set: */
		if (!something_changed)
			old_clock = *clock;
		something_changed = 1;
	}

	/* if something changed, recalculate the ntp adjustment value: */
	if (something_changed) {
		/* accumulate current leftover cycles using old_clock: */
		if (cycle_delta) {
			delta_nsec = cyc2ns_rem(&old_clock, ntp_adj,
						cycle_delta, &remainder);
			cycle_last = cycle_now;
			__increment_system_time(delta_nsec);
			ntp_advance(delta_nsec);
		}

		/* recalculate the ntp adjustment and fixed interval values: */
		ppm_last = ppm;
		ntp_adj = ppm_to_mult_adj(clock, ppm);
		ts_interval = calculate_clocksource_interval(clock, ntp_adj,
					INTERVAL_LEN);
	}

	update_legacy_time_values();

	write_sequnlock_irqrestore(&system_time_lock, flags);

	if (clocksource_changed)
		hrtimer_clock_notify();

	/* notify the posix timers if wall_time_offset changed */
	if (leapsecond)
		clock_was_set();

	/* set us up to go off on the next interval: */
	mod_timer(&timeofday_timer,
		jiffies + 1 + msecs_to_jiffies(PERIODIC_INTERVAL_MS));
}

/**
 * timeofday_is_continuous - check to see if timekeeping is free running
 */
int timeofday_is_continuous(void)
{
	unsigned long seq;
	int ret;

	do {
		seq = read_seqbegin(&system_time_lock);

		ret = clock->is_continuous;

	} while (read_seqretry(&system_time_lock, seq));

	return ret;
}

/**
 * timeofday_init - Initializes time variables
 */
void __init timeofday_init(void)
{
	unsigned long flags;

	write_seqlock_irqsave(&system_time_lock, flags);

	/* initialize the clock variable: */
	clock = get_next_clocksource();

	/* initialize cycle_last offset base: */
	cycle_last = read_clocksource(clock);

	/* initialize wall_time_offset to now: */
	/* XXX - this should be something like ns_to_ktime() */
	wall_time_offset = ktime_add_ns(wall_time_offset,
					read_persistent_clock());

	/* initialize timespec values: */
	wall_time_ts = ktime_to_timespec(ktime_add(system_time,
						wall_time_offset));
	monotonic_time_offset_ts = ktime_to_timespec(wall_time_offset);

	/* clear NTP scaling factor & state machine: */
	ntp_adj = 0;
	ntp_clear();
	ts_interval = calculate_clocksource_interval(clock, ntp_adj,
				INTERVAL_LEN);

	/* initialize legacy time values: */
	update_legacy_time_values();

	write_sequnlock_irqrestore(&system_time_lock, flags);

	/* install timeofday_periodic_hook timer: */
	init_timer(&timeofday_timer);
	timeofday_timer.function = timeofday_periodic_hook;
	timeofday_timer.expires = jiffies + 1
				+ msecs_to_jiffies(PERIODIC_INTERVAL_MS);
	add_timer(&timeofday_timer);
}
